/*
 Java Class Inspector
   Copyright (C) 1996 Clark VERBRUGGE.     Contact: clump@cs.mcgill.ca
   Copyright (C) 1996 McGill University. 
   Copyright (C) 1996 McCAT System Group.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA. */

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

public class CFG {

	method_info method;         // CFG is for this method
	BasicBlock cfg;             // actual cfg is a graph of BasicBlocks
	Hashtable h;                // for associating leaders (Instructions) with BBs

	CFG(method_info m) {
		Instruction i,head;
		BasicBlock bb,blast;
		method = m;

		h = new Hashtable(100,25);
		i = buildBasicBlock(m.instructions);
		cfg = new BasicBlock(m.instructions);
		blast = cfg;
		h.put(m.instructions,cfg);
		while (i != null) {
			 head = buildBasicBlock(i);
			 bb = new BasicBlock(i);
			 blast.next = bb;
			 blast = bb;
			 h.put(i,bb);
			 i = head;
		}
		//buildCFG();
	}

	public void showCFG(String methodname) {
		BasicBlock b,bb;
		String s=null;
		System.out.println("Basic block list for: " + methodname);
		b = cfg;
		while (b!=null) {
			/*for (Enumeration e = h.elements();e.hasMoreElements();) {
			  b = (BasicBlock)(e.nextElement());*/
			System.out.println("Basic block for " + b.head.label + " to " +
							   b.tail.label);
			System.out.print("Predecessors:");
			for (Enumeration e = b.pred.elements();e.hasMoreElements();) {
				bb = (BasicBlock)(e.nextElement());
				if (bb!=null) s = Integer.toString(bb.head.label);
				else s = "<call>";
				System.out.print(" [" + s + "]");
			}
			System.out.println("");
			System.out.print("Successors:");
			for (Enumeration e = b.succ.elements();e.hasMoreElements();) {
				bb = (BasicBlock)(e.nextElement());
				if (bb!=null) s = Integer.toString(bb.head.label);
				else s = "<call>";
				System.out.print(" [" + s + "]");
			}
			System.out.println("");
			  
			ByteCode.showCode(b.head,b.head.label);
			b = b.next;
		}
	}

	// Constructs the actual control flow graph. Assumes the hash table
	// currently associates leaders with BasicBlocks, this function
	// builds the next[] and prev[] pointer arrays.
	public void buildCFG() {
		BasicBlock b,bb;
		Instruction i,branches[],nexti;
		int numb;
		b = cfg;
		System.out.println("Building CFG...");
		while (b!=null) {
			/*for (Enumeration e = h.elements();e.hasMoreElements();) {
			  b = (BasicBlock)(e.nextElement());*/
			i = b.tail;
			if (i.branches) {
				// must look out for athrow, which can call an exception handler
				if (i instanceof Instruction_Athrow) {
					Code_attribute ca = ClassFile.locate_code_attribute(method);
					// see how many targets it can reach.  Note that this is a
					// subset of the exception_table.
					int icount = 1;
					// not quite a subset---could also be that control exits this
					// method, so start icount at 1
					for (int k = 0; k<ca.exception_table_length;k++) {
						if (i.label >= ca.exception_table[k].start_inst.label &&
						    (ca.exception_table[k].end_inst==null || 
						     i.label < ca.exception_table[k].end_inst.label)) {
							icount++;
						}
					}
					branches = new Instruction[icount];
					branches[0] = null;
					icount = 1;
					for (int k = 0; k<ca.exception_table_length;k++) {
						if (i.label >= ca.exception_table[k].start_inst.label &&
						    (ca.exception_table[k].end_inst==null || 
						     i.label < ca.exception_table[k].end_inst.label)) {
							branches[icount] = ca.exception_table[k].handler_inst;
						}
					}
				} else {
					nexti = (i.next==null) ? ((b.next==null) ? null : b.next.head) : i.next;
					branches = i.branchpoints(nexti);
				}
				if (i.calls) numb = 1;
				else numb = 0;
				if (branches!=null)
				  numb += branches.length;
				b.succ.ensureCapacity(b.succ.size()+numb);
				// System.out.println(i.label + "(" + i + " has " + numb + " branches)");
				if (i.calls && b.next!=null) {
					b.succ.addElement(b.next);
					b.next.pred.addElement(b);
				}
				if (branches!=null) {
					int j;
					for (j=0;j<branches.length;j++) {
						if (branches[j]!=null) {
							bb = (BasicBlock)(h.get(branches[j]));
							if (bb==null)
							  System.out.println("Warning: target of a branch is null");
							else {
								b.succ.addElement(bb);
								bb.pred.addElement(b);
							}
						}
					}
				}
			} else if (b.next!=null) { // BB ended not with a branch, so just go to next
				b.succ.addElement(b.next);
				b.next.pred.addElement(b);
			}
			b = b.next;
		}
	}

	// given the list of instructions head, this pulls off the front
	// basic block, terminates it with a null, and returns the next
	// instruction after.
	public static Instruction buildBasicBlock(Instruction head) {
		Instruction newhead,i;
		i = head;
		while (i!=null) {
			if (i.branches || (i.next!=null && i.next.labelled)) {
				newhead = i.next;
				i.next = null;
				return newhead;
			}
			i = i.next;
		}
		return null;
	}
}
