/*
 Java Class Inspector
   Copyright (C) 1996 Clark VERBRUGGE.     Contact: clump@cs.mcgill.ca
   Copyright (C) 1996 McGill University. 
   Copyright (C) 1996 McCAT System Group.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA. */

import java.io.*;

public class Coffi {

	public static final String VERSIONSTRING = "Version 0.10";

	static String fn;
	static FileInputStream f;                  // for reading in a class file
	static DataInputStream d;

	static FileOutputStream ff;                // for writing a class file
	static DataOutputStream dd;

	// Returns true on success, false on failure
	public static boolean open(boolean forwrite) {
		try {
			if (forwrite) 
				ff = new FileOutputStream(fn);
			else
				f = new FileInputStream(fn);
		} catch(FileNotFoundException e) {
			if (fn.indexOf(".class")>=0) {
				System.out.println("Can't find " + fn);
				return false;
			}
			fn = fn + ".class";
			try {
				if (forwrite) 
					ff = new FileOutputStream(fn);
				else
					f = new FileInputStream(fn);
			} catch(FileNotFoundException ee) {
				System.out.println("Can't find " + fn);
				return false;
			} catch(IOException ee) {
				System.out.println("IOException with " + fn + ": " + ee.getMessage());
				return false;
			}
		} catch(IOException e) {
			System.out.println("IOException with " + fn + ": " + e.getMessage());
			return false;
		}
		if (forwrite) {
			dd = new DataOutputStream(ff);
			if (dd==null) return false;
		} else {
			d = new DataInputStream(f);
			if (d==null) return false;
		}
		return true;
	}

	public static boolean close() {
		try {
			if (d!=null)
				d.close();
			if (f!=null)
				f.close();
			if (dd!=null)
				dd.close();
			if (ff!=null)
				ff.close();
			dd = null; ff = null; d = null; f = null;
		} catch(IOException e) {
			System.out.println("IOException with " + fn + ": " + e.getMessage());
			return false;
		}
		return true;
	}

	public static void main(String argv[]) {
		ClassFile cf = null;

		if (argv.length > 0) {
			int i = 0;
			while (i < argv.length) {
				if ("-m".compareTo(argv[i])==0 && i<argv.length-1) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.showMethod(argv[++i]);
				} else if ("-g".compareTo(argv[i])==0 && i<argv.length-1) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.showCFG(argv[++i]);
				} else if ("-h".compareTo(argv[i])==0) {
					System.out.println("java ClassFile [arguments]");
					System.out.println(" where arguments are processed in left->right order, " +
					                   "and can be:");
					System.out.println(" [-f] <file[.class]> Open and parse the given class file");
					System.out.println("                     (use -f if file begins with a '-')");
					System.out.println("  -m <method>        Displays code for method <method>");
					System.out.println("  -g <method>        Show CFG for method <method>");
					System.out.println("  -lm                List all methods");
					System.out.println("  -lf                List all fields");
					System.out.println("  -lc                List constant pool");
					System.out.println("  -t                 Run test routine");
					System.out.println("  -w                 Recompile and save code");
					System.out.println("  -v                 Display version and exit");
					System.out.println("  -h                 This help");
				} else if ("-v".compareTo(argv[i])==0) {
					System.out.println(VERSIONSTRING);
					System.exit(0);
				} else if ("-t".compareTo(argv[i])==0) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.testit();
				} else if ("-lm".compareTo(argv[i])==0) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.listMethods();
				} else if ("-lc".compareTo(argv[i])==0) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.listConstantPool();
				} else if ("-lf".compareTo(argv[i])==0) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else
						cf.listFields();
				} else if ("-w".compareTo(argv[i])==0) {
					if (cf==null) System.out.println("Classfile arg must precede work on it!");
					else {
						cf.recompileAll();
						//cf.addField(ACC_PUBLIC,"hey_im_a_new_field",DESC_INT);
						//System.out.println("Writing " + argv[0]);
						if (open(true)) {
							cf.writeClass(dd);
							close();
						}
					}
				} else {
					if ("-f".compareTo(argv[i])==0 && i<argv.length-1) {
						i++;
					} else if(argv[i].charAt(0)=='-') {
						System.out.println("Unrecognized option: " + argv[i]);
						System.out.println("try -h for help");
						System.exit(0);
					}
					if (cf!=null) {
					  System.out.println("Hey, only 1 classfile at a time!");
					} else {
					    fn = argv[i];
						cf = new ClassFile(fn);
						//cf.debug = true;
						if (open(false)) {
							System.out.println("Reading " + fn);
							cf.read(d);
							close();
						}
						cf.parse();        // parse all methods
					}
				}
				i++;
			}
		}	
		// System.out.println("Done");
	}
}
